/*
 *  YASS 1.10
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, 
 *  Christophe Valmir, Steven Corroy
 *
 *  noe|<A>|lifl.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */


#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <string.h>
#include <time.h>

#include "util.h"
#include "global_var.h"
#include "kword.h"
#include "tuple.h"
#include "assemble.h"
#include "prdyn.h"
#include "proba.h"
#include "display.h"
#include "regroup.h"
#include "threads.h"

static void usage()
{
    SETCOLOR(stderr,GREEN);
    fprintf(stderr, "*  usage:\n");
    RESET(stderr);
    fprintf(stderr, "  " PACKAGE_NAME " [options] { fasta.fas | fasta.fas multifasta.fas }\n");
    
    fprintf(stderr, "      -h       display this Help screen\n");
    fprintf(stderr,
	    "      -d <int>    0 : Display alignment positions (default)\n"
	    "                  1 : Display alignment positions + alignments + statistics\n"
	    "                  2 : Display blast-like tabular output\n"
	    "                  3 : Display light tabular output (better for post-processing)\n");

    fprintf(stderr,
	    "      -r <int>    0 : process forward (query) strain\n"
	    "                  1 : process Reverse complement strain\n"
	    "                  2 : process both forward and Reverse complement strains (default)\n");
    fprintf(stderr, "      -o <str> Output file\n");
    fprintf(stderr, "      -v       display the current Version\n");
    fprintf(stderr,
	    "                                                                               \n");
    fprintf(stderr,
	    "      -M <int> select a scoring Matrix (3 by default):\n"
	    "               [Match,Transversion,Transition],(Gopen,Gext)\n"
	    "               0 : [1,-3,-2],( -8,-2)   1 : [3,-2,-1],(-12,-4)\n"
	    "               2 : [3,-4,-1],(-16,-4)   3 : [5,-4,-2],(-16,-4)\n");
    fprintf(stderr,
	    "      -C <int>[,<int>[,<int>[,<int>]]] \n"
	    "               reset match/mismatch/transistion/other Costs (penalties)\n");
    fprintf(stderr,
	    "      -G <int>,<int> reset Gap opening/extension penalties\n");
    fprintf(stderr,
	    "      -L <real>,<real> reset Lambda and K parameters of Gumbel law\n");
    fprintf(stderr, "      -X <int> Xdrop threshold score (default %i)\n",
	    gp_xdrop);
    fprintf(stderr, "      -E <int> E-value threshold (default %1.2g)\n",
    gp_expectation_value);
    fprintf(stderr,
	    "                                                                               \n");
    fprintf(stderr,
	    "      -O <int> limit number of Output alignments (default %i)\n",
	    gp_nbmaxlines);
    fprintf(stderr,
	    "      -S <int> Select query sequence from multi-fasta query file (default %i)\n",
	    gp_selection_fasta);
    fprintf(stderr,
	    "      -T <int> forbid aligning too close regions (e.g. Tandem repeats)\n"
	    "               valid for single sequence comparison only (default %i bp)\n",
	    gp_distdiag);
    fprintf(stderr,
	    "                                                                               \n");
    fprintf(stderr,
	    "      -s <int> group Size              (default %i bp)\n",
	    gp_minimum_score_seeds);
    fprintf(stderr,
	    "      -p <str> seed Pattern                                                    \n"
	    "                 * use \'#\' for match \n"
	    "                 * use \'@\' for match or transition \n"
	    "                 * use \'_\' or \'-\' for joker (recommended)\n"
	    "                 * use \',\' for separator (max: %d seeds)\n"
	    "                 - example 1: \n"
	    "                    " PACKAGE_NAME " file.fas -p  \"#@#__##__#_##@#\" \n"
	    "                 - example 2: \n"
	    "                    " PACKAGE_NAME " file.fas -p \"##_#_#@#_##@_##,##@#__@__##_#__###\" \n",
	    MAX_SEED);
    fprintf(stderr,
	    "                                                                               \n");
    
    fprintf(stderr,
	    "      -e <real> minimal allowed Entropy of trinucleotide distribution  \n"
	    "                ranging between 0 (no filter) and 6, default %1.2f \n",
	    gp_entropy_min);
    fprintf(stderr,
	    "      -t <real> Trim out over-represented seeds codes                  \n"
	    "               ranging between 0.0 (no trim) and +inf, default %1.2f \n",
	    gp_t);

    fprintf(stderr,
	    "      -a <int> statistical tolerance Alpha (%%) (default %i%%)\n",
	    gp_alpha_percent);
    fprintf(stderr,
	    "      -i <int> Indel rate (%%)                  (default %i%%)\n",
	    gp_indels_percent);
    fprintf(stderr,
	    "      -m <int> Mutation rate (%%)               (default %i%%)\n",
	    gp_mutations_percent);
    fprintf(stderr,
	    "                                                                               \n");
    fprintf(stderr,
	    "      -W <int>,<int> Window <min,max> range for post-processing and grouping\n"
	    "                     alignments (default <%i,%i>) \n",
	    gp_win_min, gp_win_max);
    fprintf(stderr,
	    "      -w <real> Window size coefficient for post-processing and grouping        \n"
	    "               alignments (default %g). NOTE : -w 0 disables post-processing\n\n",
	    gp_win_mul);

    
#ifdef LOW_MEMORY
    fprintf(stderr,
	    "      -Hsize <int> Hash table size (default %i)\n" 
	    "                   (high values will speed up " PACKAGE_NAME ")\n",
	    gp_hash_size);
    fprintf(stderr,
	    "      -Helem <int> Hash table number of elements (default %i) \n" 
	    "                   (the higher, the better for " PACKAGE_NAME " to detect alignements)\n",
	    gp_hash_element);

#endif

#ifdef MEM_ALLOCATED
    fprintf(stderr,
	    "      -Alloc <int> maximal total Allocation allowed (default %lu) \n",
	    gp_max_mem_allocated);
#endif

    RESET(stderr);
    RESET(stdout);
    exit(-1);
}

static void vers()
{
    fprintf(stderr, "version " PACKAGE_VERSION "\n");
    exit(-1);
}



static void error(char *E1, char *E2, char *E3 )
{
    SETCOLOR(stderr, RED);
    fprintf(stderr, "** ERROR:\n  ");
    RESET(stderr);
    fprintf(stderr,"%s", E1);
    fprintf(stderr,"%s", E2);
    fprintf(stderr,"%s", E3);
    fprintf(stderr, "\n");
    usage();
}


/*
 * read several seed patterns separated with ','
 */
static void parsepattern(int argc, char ** argv, int * i) {
  char * s_begin = NULL;
  char * s_end   = NULL;
  (*i)++;
  if (*i >= argc) 
    error("\"",argv[*i-1],"\" found without motif");
  if (!(strchr(argv[*i],'#')||strchr(argv[*i],'@')))        
    error(" pattern \"",argv[*i],"\"  doesn't have any \"#\" or \"@\" symbol");
  
  gp_nb_seeds = 0;
  s_begin = argv[*i];
  while(1) {
    if(gp_nb_seeds == MAX_SEED)       
      error(" pattern \"",argv[*i-1],"\" has too many seeds");
    
    s_end = strchr(s_begin,',');
    if (s_end) {
      *s_end = '\0';
      gp_motifs[gp_nb_seeds++] = (char *) strdup(s_begin);
      s_begin = s_end + 1;
    } else {
      gp_motifs[gp_nb_seeds++] = (char *) strdup(s_begin);
      break;
    }
  }
}

/*
 * read the output filename
 */
static void parseoutfile(int argc, char ** argv, int * i) {
  (*i)++;
  if ((*i) >= argc)  
    error("\"",argv[(*i)-1],"\" found without filename");
  gv_outstream = fopen(argv[*i],"w");
  if (!gv_outstream) 
    error("Can't create file \"",argv[*i],"\"");
}

/*
 * read one integer -<Option> <int>
 */
static void parseint(int argc, char ** argv, int * i, int * var, int min, int max) {
  int i_tmp;
  (*i)++;
  if ((*i)>=argc)                                                                 
    error("\"",argv[(*i)-1],"\" found without argument");
  i_tmp = strtol(argv[(*i)],NULL,10);
  if (errno != 0 )        
    error("\"",argv[(*i)-1],"\" is not followed by an integer or by an integer outside the range");
  if (i_tmp < min || i_tmp  > max)
    error("\"",argv[(*i)-1],"\" is followed by an integer outside the range");
  *var = i_tmp;
}


/*
 * read one unsigned long integer -<Option> <int>
 */
static void parseulint(int argc, char ** argv, int * i, unsigned long int * var, unsigned long int min, unsigned long int max) {
  unsigned long int uli_tmp;
  (*i)++;
  if ((*i)>=argc)                                                                 
    error("\"",argv[(*i)-1],"\" found without argument");
  uli_tmp = strtoul(argv[(*i)],NULL,10);
  if (errno != 0 )        
    error("\"",argv[(*i)-1],"\" is not followed by an integer or by an integer outside the range");
  if (uli_tmp < min || uli_tmp > max)
    error("\"",argv[(*i)-1],"\" is followed by an integer outside the range");
  *var = uli_tmp;
}


/*
 * read two integers -<Option> <int>,<int>
 */
static void parsedint(int argc, char ** argv, int * i, int * var1, int min1, int max1, int * var2, int min2, int max2) {
  int i_tmp1, i_tmp2, err;
  (*i)++;
  if ((*i)>=argc)                                                                  
    error("\"",argv[(*i)-1],"\" found without argument");
  err = sscanf(argv[(*i)],"%i,%i",&i_tmp1,&i_tmp2);
  if (err != 2 )                   
    error("\"",argv[(*i)-1],"\" is not followed by an <int>,<int>");
  if (i_tmp1 < min1 || i_tmp1 > max1 || i_tmp2 < min2 || i_tmp2 > max2)                    
    error("\"",argv[(*i)-1],"\" is followed by integers  outside the range");
  *var1 = i_tmp1;
  *var2 = i_tmp2;
}


static void parsedouble(int argc, char ** argv, int * i, double * var, double min, double max) {
  double d_tmp;
  (*i)++;
  if ((*i)>=argc)                                                                  
    error("\"",argv[(*i)-1],"\" found without argument");
  d_tmp = strtod(argv[(*i)],NULL);
  if (errno != 0 )                                                                 
    error("\"",argv[(*i)-1],"\" is not followed by a double") ;
  if (d_tmp>max || d_tmp <min)                                                       
    error("\"",argv[(*i)-1],"\" is followed by a double outside the range");
  *var = d_tmp;
}

static void parseddouble(int argc, char ** argv, int * i, double * var1, double min1, double max1, double * var2, double min2, double max2) {
  double d_tmp1, d_tmp2;
  int    err;
  (*i)++;
  if (*i>=argc)
    error("\"",argv[(*i)-1],"\" found without argument");
  err = sscanf(argv[*i],"%lf,%lf",&d_tmp1,&d_tmp2);
  if (err != 2 )
    error("\"",argv[(*i)-1],"\" is not followed by an <real>,<real>");
  if (d_tmp1 < min1 || d_tmp1 > max1 || d_tmp2 < min2 || d_tmp2 > max2)
    error("\"",argv[*i-1],"\" is followed by integers outside the range");
  *var1 = d_tmp1;
  *var2 = d_tmp2;
}



/*
 * read matrix substitution costs  -<Option> <int> ( ,<int>  ( ,<int>  ( ,<int> )))
 */
static void parsecost(int argc, char ** argv, int * i) {
  char * s_begin,* s_end;
  int  cost_granularity = 0;
  (*i)++;
  if ((*i)>=argc)                                             
    error("\"",argv[(*i)-1],"\" found without argument");

  /* 1) read parameters */
  s_begin = argv[(*i)];
  cost_granularity = 0;

  while(1) {
    
    if(cost_granularity == 4)       
      error(" -\"",argv[*i-1],"\" has too many arguments");
    
    s_end = strchr(s_begin,',');
    gp_costs[cost_granularity++] = strtol(s_begin,NULL,10);
    
    if (errno != 0 )
      error("\"",argv[(*i)-1],"\" is not followed by a list of integers");

    if (s_end)
      s_begin = s_end + 1;
    else
      break;
  }
  
  /* 2) warnings on positives/negatives values */
  if (gp_costs[0] < 0){
    _WARNING(" " PACKAGE_NAME " converts match cost into positive value");
    gp_costs[0] =   ABS(gp_costs[0]);
  }
  
  if ( gp_costs[1] > 0 || gp_costs[2] > 0 || gp_costs[3] > 0 ){
    _WARNING(" " PACKAGE_NAME " converts mismatch costs (transition/transversion/other) into negative values");
    gp_costs[1] = - ABS(gp_costs[1]);
    gp_costs[2] = - ABS(gp_costs[2]);
    gp_costs[3] = - ABS(gp_costs[3]);
  }

  /* 3) affect costs depending on parameters granularity */
  switch (cost_granularity){
  case 1: /* only match cost given : other costs are -match */
    gp_costs[1] =  gp_costs[2] = gp_costs[3] = - gp_costs[0];
    break;
  case 2: /* match/mismatch costs : other costs are mismatches */
    gp_costs[2] = gp_costs[3] = gp_costs[1];
    break;
  case 3: /* match/mismatch/transition : other is ~ to mismatch */
    gp_costs[3] = gp_costs[1];
    break;
  }
}


/*
 * read complete parameters (unless for files)
 */

static void ParseParams(int argc, char ** argv, int * i) {
  if(!strcmp(argv[*i],"-a")||!strcmp(argv[*i],"--alpha"))
    parseint(argc,argv,i,
	      &gp_alpha_percent,1,99);
  else if(!strcmp(argv[*i],"-i")||!strcmp(argv[*i],"--indels"))
    parseint(argc,argv,i,
	     &gp_indels_percent,0,100);
  else if(!strcmp(argv[*i],"-m")||!strcmp(argv[*i],"--mutations"))
    parseint(argc,argv,i,
	     &gp_mutations_percent,0,99);
  else if(!strcmp(argv[*i],"-s")||!strcmp(argv[*i],"--minimumscoreseeds"))
    parseint(argc,argv,i,
	     &gp_minimum_score_seeds,0,0x7fffffff);
  else if(!strcmp(argv[*i],"-M")||!strcmp(argv[*i],"--Matrix")){
    parseint(argc,argv,i,
	     &gp_matrix,0,3);
    gp_costs[0]            = SUBMATRIXTABLE[gp_matrix][0];
    gp_costs[1]            = SUBMATRIXTABLE[gp_matrix][1];
    gp_costs[2]            = SUBMATRIXTABLE[gp_matrix][2];
    gp_costs[3]            = SUBMATRIXTABLE[gp_matrix][3];  
    gp_cost_gap_opened     =    INDELSTABLE[gp_matrix][0];
    gp_cost_gap_continued  =    INDELSTABLE[gp_matrix][1];
  }
  else if(!strcmp(argv[*i],"-T")||!strcmp(argv[*i],"--Tandems"))
    parseint(argc,argv,i,
	     &gp_distdiag,0,0x7fffffff);
  else if(!strcmp(argv[*i],"-X")||!strcmp(argv[*i],"--Xdrop")){
    parseint(argc,argv,i,
	     &gp_xdrop,0,100);
  }
  
  else if(!strcmp(argv[*i],"-O")||!strcmp(argv[*i],"--nbO"))
    parseint(argc,argv,i,
	     &gp_nbmaxlines,1,0x7fffffff);
  else if(!strcmp(argv[*i],"-S")||!strcmp(argv[*i],"--nbs"))
    parseint(argc,argv,i,
	     &gp_selection_fasta,1,0x7fffffff);
  else if(!strcmp(argv[*i],"-G")||!strcmp(argv[*i],"--Gaps")){
    parsedint(argc,argv,i,
	      &gp_cost_gap_opened   ,-100000,+100000,
	      &gp_cost_gap_continued,-100000,+100000);
    if( gp_cost_gap_opened > 0 ||  gp_cost_gap_continued > 0){
      _WARNING(" " PACKAGE_NAME " converts  gaps costs into negative values");
      gp_cost_gap_opened     = -ABS(gp_cost_gap_opened);
      gp_cost_gap_continued  = -ABS(gp_cost_gap_continued);
    }
    
  }
  
  else if(!strcmp(argv[*i],"-C")||!strcmp(argv[*i],"--Costs"))
    parsecost(argc,argv,i);
  else if(!strcmp(argv[*i],"-W")||!strcmp(argv[*i],"--Windows"))
    parsedint(argc,argv,i,
	      &gp_win_min,0,0x7fffffff,
	      &gp_win_max,0,0x7fffffff);
  else if(!strcmp(argv[*i],"-w")||!strcmp(argv[*i],"--window"))
    parsedouble(argc,argv,i,
	        &gp_win_mul,0,1e+16);
  else if(!strcmp(argv[*i],"-E")||!strcmp(argv[*i],"--Expectation"))
    parsedouble(argc,argv,i,
		&gp_expectation_value,1e-100,1e+10);
  else if(!strcmp(argv[*i],"-L")||!strcmp(argv[*i],"--Lambda"))
    parseddouble(argc,argv,i,
		 &gp_lambda_blast,1e-9,1e+9,		 
		 &gp_k_blast,     1e-9,1e+9);
  else if(!strcmp(argv[*i],"-r")||!strcmp(argv[*i],"--reverse"))
    parseint(argc,argv,i,
	     &gp_reverse,0,2);
  else if(!strcmp(argv[*i],"-d")||!strcmp(argv[*i],"--display"))
    parseint(argc,argv,i,
	     &gp_display,0,3);
  else if(!strcmp(argv[*i],"-p")||!strcmp(argv[*i],"--pattern"))
    parsepattern(argc,argv,i);
  else if(!strcmp(argv[*i],"-e")||!strcmp(argv[*i],"--entropy"))
    parsedouble(argc,argv,i,
		&gp_entropy_min,0.0,6.0);
  else if(!strcmp(argv[*i],"-t")||!strcmp(argv[*i],"--trim"))
    parsedouble(argc,argv,i,
	     &gp_t,0.0,1e+100);
  else if(!strcmp(argv[*i],"-o")||!strcmp(argv[*i],"--output"))
    parseoutfile(argc,argv,i);
  else if(!strcmp(argv[*i],"-h")||!strcmp(argv[*i],"--help"))
    usage();
  else if(!strcmp(argv[*i],"-v")||!strcmp(argv[*i],"--version"))
    vers();
  else if(!strcmp(argv[*i],"-Helem"))
    parseint(argc,argv,i,
	     &gp_hash_element,0,~0);
  else if(!strcmp(argv[*i],"-Hsize"))
    parseint(argc,argv,i,
	     &gp_hash_size,0,~0);
#ifdef MEM_ALLOCATED
  else if(!strcmp(argv[*i],"-Alloc"))
    parseulint(argc,argv,i,
	       &gp_max_mem_allocated,0,~0);
#endif	
  else 
    error("\"",argv[*i],"\" is not a valid argument");
}


/*
 * read filenames
 */
static void ParseName(int argc, char ** argv, int * i) {
  FILE * f = NULL;
  if (gp_nbfiles == 2)
    error("Too many files :\"",argv[*i],"\" is the third one ...");
  strncpy(gp_files[gp_nbfiles],argv[*i],255);
  if (!(f=fopen(gp_files[gp_nbfiles],"r")))
    error("Can't open file\"",argv[*i],"\"");
  fclose(f);
  gp_nbfiles++;
}

/*
 * scan arguments function
 */

static void ScanArg(int argc, char ** argv){
  int i,x,y;
  for(i=1;i<argc;i++){
    if(argv[i][0] == '-' ){
      ParseParams(argc,argv,&i);
    }else{
      ParseName(argc,argv,&i);
    }
  }
  
  /* 1) parameters post-checking */
  if (gp_nbfiles == 0)
    error("No file found"," ... ","give at least one file ...");
  /* FIXME */
  /*  if ( gp_seeds_bitweight[0] >= 2*gp_minimum_score_seeds ) 
   *   error("weight needs to be under s","",".");
   */


  /* 2) fill in seeds and automata */
  ComputeLength();
  AutomataOverlaps(AUTOMATADELPH);
  
  /* 3) fill-in substitution matrix */
  for (x=0;x<16;x++){
    for (y=0;y<16;y++){      
      if (x<6 && y<6)
	if (x==y)
	  gp_substitution_matrix[x][y] = gp_costs[0];
	else 
	  if ((x%2) == (y%2))
	    gp_substitution_matrix[x][y]= gp_costs[2];
	  else
	    gp_substitution_matrix[x][y]= gp_costs[1]  ;
      else
	gp_substitution_matrix[x][y] = gp_costs[1];
      
      gp_cost_max_substitution_matrix =  MAX(
					     ABS(gp_substitution_matrix[x][y]),
					     gp_cost_max_substitution_matrix
					     );
      
      /* symbols U (6) ~ T (1) */
      if (y==6)
	gp_substitution_matrix[x][y]=gp_substitution_matrix[x][1];
      if (x==6)
	gp_substitution_matrix[x][y]=gp_substitution_matrix[1][y];
      if (y==6 && x==6)
	gp_substitution_matrix[x][y]=gp_substitution_matrix[1][1];
    }
  }
}





/*
 *   Function to be lauched by threads
 */

static
#ifdef  BIPROC
#ifdef WIN32
  DWORD WINAPI forward(PVOID parameters_assemble)
#else
  void * forward(void * parameters_assemble)
#endif	
     
#else
  void * forward(void * parameters_assemble)
#endif
{
   
  /*
   * 1) Assemble
   */

  /* global variable for one thread */
  Feature * f = parameters_assemble;
  
  initialise_alignment(8 * MEGA,f);

  CreateCountMA(f);
  
  if (gp_nbfiles == 2) {
    MultiAssemble_Double(f->chunk_query, gp_querysize_chunk,
			 gp_text, gp_textsize, gp_nbchunks_text,
			 gp_chunkname_text, gp_chunksize_text,
			 gp_chunkstrt_text, f);
 
  } else {
    if (gp_nbchunks_query > 1) {
      /* FIXME : need  MultiAssemble_SingleRev which do no report comparison on the same diagonal neither the half top matrix */
      MultiAssemble_Double(f->chunk_query, gp_querysize_chunk,
			   gp_text, gp_textsize, gp_nbchunks_text,
			   gp_chunkname_text, gp_chunksize_text,
			   gp_chunkstrt_text, f);
    } else {
      Assemble_Single(f->chunk_query, gp_querysize_chunk, f);
    }
  }


#ifdef DEBUG_MEMORY
  fprintf (stderr,"Memorized MA size before post-processing :\n");SizeMA (f->first_MA);
#endif 
  
  /* 
   * 2) Post-processing 
   */  
#ifdef STATS
  last_clock = clock();
#endif

  /*
   * 2.1) Filtering 	 
   */
  
  DISPLAY_BEGIN(filtering_mutex_begin,filtering_begin ,"\n\n filtering   ...  ");
  if (f->first_MA)
    EntropyFilter_MAList(f,
			 f->chunk_query,
			 gp_text, gp_chunkstrt_text);
  DISPLAY_END(filtering_mutex_end,filtering_end);

  /* 
   * 2.2) MA regrouping
   */

  DISPLAY_BEGIN(regroup_mutex_begin,regroup_begin ," grouping   ....  ");
  if (f->first_MA)
    main_multi_regroup(f);
  DISPLAY_END(regroup_mutex_end,regroup_end);
  

#ifdef STATS
  f->clock_post += clock() - last_clock;
#endif
 
  END_THREAD();
}







/************************
 * HOSTILITY STARTS HERE *
 ************************/

int main(int argc, char *argv[]) {
    double *probability;	/* mutational even probability */
    MA * ma;
    Feature * f;

    /*[0] Default parameters */

    /* with one seed
       gp_motifs[0]=(char *)strdup("#@#__##__#_##@#");
       gp_nb_seeds=1;
    

       with 2 seeds 
       
       (Bernouilli model): 2 seeds (comment)
       1)0.750715:	#@_##_##_#__@_###	#_##@___##___#___#@#_#  (good seed, excepted for invertebral)
       2)0.753335:	#_###@#__#_#_@##	###@_#_#______@#_#__##  (average seed)
       3)0.751042:	##_#_#@#_#__@###	#_##@__#____#__##@##    (nice seeds for all sequences)

    */
    gp_motifs[0] = (char *) strdup("#@_##_##_#__@_###");
    gp_motifs[1] = (char *) strdup("#_##@___##___#___#@#_#"); 
    gp_nb_seeds = 2;

    
    /* Init semaphores */
    INIT_MUTEX();

    /* [1] command line read */
    ScanArg(argc,argv);
       
    /* [2] initialize parameters and statistics  */
    gp_mutations = (double) gp_mutations_percent / 100.00;
    gp_indels    = (double) gp_indels_percent    / 100.00;
    gp_alpha     = (double) gp_alpha_percent     / 100.00;
    gp_rho_stat =  statistical_bound_of_waiting_time2(
						      (1 - gp_mutations),
						      gp_seeds_bitweight[0] / 2,
						      gp_alpha
						      );
    gp_delta_stat = statistical_bound_of_randomwalk2(gp_indels, 
						     gp_rho_stat, 
						     gp_alpha);

    /* [3] chaining algorithm memory required */
    initialise_deltashift();

    /* [4] read files and set their attributes */
    if (gp_nbfiles == 2) {
	CreateData(gp_files[0], &gp_query, &gp_querysize, &gp_nbchunks_query,
		   &gp_chunkname_query, &gp_chunksize_query, &gp_chunkstrt_query, gp_nb_letters[0]);

	CreateData(gp_files[1], &gp_text, &gp_textsize, &gp_nbchunks_text,
		   &gp_chunkname_text, &gp_chunksize_text, &gp_chunkstrt_text, gp_nb_letters[1]);
#ifdef DEBUG_DATA
	fprintf(stderr,"----\n");
	DisplayData(gp_files[0], gp_query, gp_querysize, gp_nbchunks_query, gp_chunkname_query, gp_chunksize_query, gp_chunkstrt_query);
	fprintf(stderr,"----\n");
	DisplayData(gp_files[1], gp_text,  gp_textsize,  gp_nbchunks_text,  gp_chunkname_text,  gp_chunksize_text,  gp_chunkstrt_text);
	fprintf(stderr,"----\n");
#endif 

    } else {
	CreateData(gp_files[0], &gp_text, &gp_textsize, &gp_nbchunks_text,
		   &gp_chunkname_text, &gp_chunksize_text, &gp_chunkstrt_text, gp_nb_letters[0]);
	
	gp_query           = gp_text;
	gp_querysize       = gp_textsize;
	gp_nbchunks_query  = gp_nbchunks_text;
	gp_chunkname_query = gp_chunkname_text;
	gp_chunksize_query = gp_chunksize_text;
	gp_chunkstrt_query = gp_chunkstrt_text;

	gp_nb_letters[1][0] = gp_nb_letters[0][0];
	gp_nb_letters[1][1] = gp_nb_letters[0][1];
	gp_nb_letters[1][2] = gp_nb_letters[0][2];
	gp_nb_letters[1][3] = gp_nb_letters[0][3];
	
#ifdef DEBUG_DATA
	fprintf(stderr,"----\n");
	DisplayData(gp_files[0], gp_query, gp_querysize, gp_nbchunks_query, gp_chunkname_query, gp_chunksize_query, gp_chunkstrt_query);
	fprintf(stderr,"----\n");
#endif 

    }

    
    if (gp_nbchunks_query < gp_selection_fasta) {
      error("selected fastachunk does not exist in\"", gp_files[0],
	    "\" file , please check your -S parameter");
    }
    gp_query_chunk     = gp_query + gp_chunkstrt_query[gp_selection_fasta - 1];
    gp_querysize_chunk = gp_chunksize_query[gp_selection_fasta - 1];
  
#ifdef DEBUG_COST
    fprintf(stderr, "costs: {match=%d, mis=%d, trans=%d, other=%d} - {gap-open:%d, gap-continue:%d}\n",
	    gp_costs[0],gp_costs[1],
	    gp_costs[2],gp_costs[3],
	    gp_cost_gap_opened,gp_cost_gap_continued
	    );
    
#endif

     
    /*
     * compute lambda and K
     */

    if (gp_lambda_blast < 0 && gp_k_blast < 0 ) {
      probability     = computeProbability(); /* FIXME */
      gp_lambda_blast = computeLambda(probability);
      gp_k_blast      = computeK     (probability,gp_lambda_blast);     
    }
    
   
    gp_minscore =
      MinScore(gp_k_blast, gp_lambda_blast,
	       gp_querysize_chunk,
	       gp_textsize, gp_expectation_value);

#ifdef DEBUG_PROBA
    fprintf(stderr, "seq 1: {#A:%d, #T:%d, #G:%d, #C:%d}\n", 
	    gp_nb_letters[0][0], gp_nb_letters[0][1], 
	    gp_nb_letters[0][2], gp_nb_letters[0][3]);
    fprintf(stderr, "seq 2: {#A:%d, #T:%d, #G:%d, #C:%d}\n", 
	    gp_nb_letters[1][0], gp_nb_letters[1][1],
	    gp_nb_letters[1][2], gp_nb_letters[1][3]);
    fprintf(stderr, "align : {pr(Match):%f,\t pr(Trans):%f,\t pr(Transv):%f,\t pr(Other):%f}\n", 
	    probability[0], probability[1], 
	    probability[2], probability[3]);
    fprintf(stderr, "Lambda: %f, K: %f \n",gp_lambda_blast,gp_k_blast);
#endif
#ifdef DEBUG_DATA
    fprintf(stderr, "querychunk_size: %d, text_size: %d\n",gp_querysize_chunk,gp_textsize);   
    fprintf(stderr, "minscore : %d\n",gp_minscore);  
#endif
#ifdef DEBUG_SEED
    DisplaySeeds();
#endif
#ifdef STATS
    gv_time_spent = time(NULL);
#endif


    /* 
     * [5] Assemble algorithms 
     */
 
    f=&gv_feature[0];
    init_Feature(f);
    
    /* [5.1] Forward Chaining */
    if (gp_reverse != 1) {
      f->chunk_query    = gp_query_chunk;
      f->reverse        = 0;
      CREATE_THREAD(f->thread,forward,f);
    }

    

    /* init reverse feature */
    f=&gv_feature[1];
    init_Feature(f);
    
    /* [5.2] Complementairy Reverse Chaining if requested */
    if (gp_reverse) {
      
      /* DataRev */
      CreateReverseComplement(gp_query_chunk, gp_querysize_chunk, &gp_queryrev_chunk);
      
      f->chunk_query    = gp_queryrev_chunk;
      f->reverse        = 1;

      initialise_alignment(8 * MEGA,f);
      
      CreateCountMA(f);
      
      if (gp_nbfiles == 2) {
	/* Assemble */
	MultiAssemble_Double(f->chunk_query, gp_querysize_chunk,
			     gp_text, gp_textsize, gp_nbchunks_text,
			     gp_chunkname_text, gp_chunksize_text,
			     gp_chunkstrt_text, f);
      } else {
	/* Assemble */
	if (gp_nbchunks_query > 1) {
	  /* FIXME : need  MultiAssemble_SingleRev which do no report  comparison on the same diagonal neither the half top matrix */
	  MultiAssemble_Double(f->chunk_query, gp_querysize_chunk,
			       gp_text, gp_textsize, gp_nbchunks_text,
			       gp_chunkname_text, gp_chunksize_text,
			       gp_chunkstrt_text, f);
	} else {
	  Assemble_SingleRev(f->chunk_query, gp_query_chunk,
			     gp_querysize_chunk, f);
	}
      }

      
#ifdef DEBUG_MEMORY
      fprintf (stderr,"Memorized MA size before post-processing :\n");SizeMA (f->first_MA);
#endif 

      
      /* [6.2] Filter */
#ifdef STATS
      last_clock = clock();
#endif
      

      
      DISPLAY_BEGIN(filtering_mutex_begin, filtering_begin ,"\n\n filtering   ...  ");
      if (f->first_MA) 
	EntropyFilter_MAList(f,
			     f->chunk_query, 
			     gp_text,gp_chunkstrt_text);
      DISPLAY_END(filtering_mutex_end, filtering_end);


      /* 
       * [6.3] Regrouping reverse MAs  
       */

      DISPLAY_BEGIN(regroup_mutex_begin,regroup_begin ," grouping   ....  ");
      if (f->first_MA)
	main_multi_regroup(f);
      DISPLAY_END(regroup_mutex_end,regroup_end);



#ifdef STATS
  f->clock_post += clock() - last_clock;
#endif

    };


    /*
     * [6.4] Waiting the child thread
     */
    
    if (gp_reverse != 1) {
      WAIT_THREAD(gv_feature[0].thread);
    }
    
    
 
    /* 
     * [6.5] Merge both MA lists 
     */
    
    gv_first_MA = gv_feature[0].first_MA;
    if (gv_first_MA == NULL) {
      gv_first_MA = gv_feature[1].first_MA;
    } else 
      gv_feature[0].last_MA->next = gv_feature[1].first_MA;
    
#ifdef DEBUG_LISTMA
       DisplayListMA (gv_first_MA);
#endif

    /*
     * result memory size
     */
    

#ifdef STATS
      last_clock = clock () ;
#endif
    /* [6.6] Sorting results */
    fprintf(stderr, " sorting   ..... ");
    if (gv_first_MA)
      ListSort_MAList(&gv_first_MA, &ma, 0x40000000);
    fprintf(stderr, " finished\n");

#ifdef STATS
    last_clock = clock () -  last_clock;
    gv_time_spent = time(NULL) - gv_time_spent;
#endif
    
    /* [7] Display Results */
    Display_Alignements(gp_querysize_chunk,
			gp_query_chunk, gp_queryrev_chunk,
			gp_chunkname_query[gp_selection_fasta - 1],
			gp_text, gp_nbchunks_text, gp_chunkname_text,
			gp_chunksize_text, gp_chunkstrt_text);
    Display_Params();
    Display_Stats(gv_feature);

    return 0;
}
